import pytest

from lasagna.planner.util import (
    NoResult,
    _extract_last_result_from_run,
    extract_is_trivial,
    extract_subtask_strings,
    extract_task_statement,
)

from lasagna.types import AgentRun
from lasagna.planner.default_prompts import PlanOutput

import copy


PLAN_OBJ = PlanOutput(
    thoughts = '... hum ...',
    task_statement = 'do the thing',
    is_trivial = False,
    subtasks = ['first thing', 'next thing', 'last thing'],
)

PLAN_DCT = PLAN_OBJ.model_dump()
assert isinstance(PLAN_DCT, dict)


NO_RESULT_RUN: AgentRun = {
    'agent': 'a',
    'type': 'parallel',
    'runs': [
        {
            'agent': 'b',
            'type': 'chain',
            'runs': [
                {
                    'agent': 'c',
                    'type': 'messages',
                    'messages': [
                        {
                            'role': 'system',
                            'text': 'do good',
                        },
                    ],
                },
            ],
        },
        {
            'agent': 'c',
            'type': 'messages',
            'messages': [
                {
                    'role': 'system',
                    'text': 'do better',
                },
            ],
        },
    ],
}


def _insert_result(result) -> AgentRun:
    run = copy.deepcopy(NO_RESULT_RUN)
    assert run['type'] == 'parallel'
    run['runs'].append({
        'agent': 'd',
        'type': 'extraction',
        'messages': [],
        'result': copy.deepcopy(result),
    })
    return run


def test_extract_last_result_from_run():
    with pytest.raises(NoResult):
        _extract_last_result_from_run(NO_RESULT_RUN)

    run = _insert_result(PLAN_OBJ)
    assert _extract_last_result_from_run(run) == PLAN_OBJ

    run = _insert_result(PLAN_DCT)
    assert _extract_last_result_from_run(run) == PLAN_DCT


def test_extract_is_trivial():
    assert extract_is_trivial(_insert_result(PLAN_OBJ)) == False
    assert extract_is_trivial(_insert_result(PLAN_DCT)) == False


def test_extract_subtask_strings():
    assert extract_subtask_strings(_insert_result(PLAN_OBJ)) == ['first thing', 'next thing', 'last thing']
    assert extract_subtask_strings(_insert_result(PLAN_DCT)) == ['first thing', 'next thing', 'last thing']


def test_extract_task_statement():
    assert extract_task_statement(_insert_result(PLAN_OBJ)) == 'do the thing'
    assert extract_task_statement(_insert_result(PLAN_DCT)) == 'do the thing'



import pytest

import asyncio
import time

from typing import List, Tuple

from lasagna.async_util import async_throttle

from lasagna.util import get_name
from lasagna.tools_util import get_tool_params


@pytest.mark.asyncio
async def test_async_throttle_max_concurrent():
    shared_decorator = async_throttle(max_concurrent=2)

    @shared_decorator
    async def foo_1() -> Tuple[str, float]:
        call_time = time.time()
        print('foo_1')
        await asyncio.sleep(0.1)
        return 'foo_1', call_time

    @shared_decorator
    async def foo_2() -> Tuple[str, float]:
        call_time = time.time()
        print('foo_2')
        await asyncio.sleep(0.1)
        return 'foo_2', call_time

    def _assert_results(expected_name: str, res: List[Tuple[str, float]]) -> None:
        res = sorted(res)  # likely already sorted, based on how the even loop schedules tasks, but that's not guaranteed, so we sort here
        assert len(res) == 10

        for name, _ in res:
            assert name == expected_name

        pairs: List[float] = []
        for i in range(0, len(res), 2):
            _, a = res[i]
            _, b = res[i+1]
            assert 0.0 <= b - a <= 0.08
            avg = (a + b) / 2
            pairs.append(avg)

        for a, b in zip(pairs, pairs[1:]):
            diff = abs(a - b)
            assert 0.09 < diff < 0.19

    tasks = [
        f()
        for _ in range(10)
        for f in [foo_1, foo_2]
    ]

    res = await asyncio.gather(*tasks)
    assert len(res) == 20

    foo_1_res = res[0::2]
    foo_2_res = res[1::2]

    _assert_results('foo_1', foo_1_res)
    _assert_results('foo_2', foo_2_res)


@pytest.mark.asyncio
async def test_async_throttle_max_per_second():
    max_per_second = 10
    shared_decorator = async_throttle(max_per_second=max_per_second)
    period = 1 / max_per_second

    @shared_decorator
    async def foo_1() -> Tuple[str, float]:
        call_time = time.time()
        print('foo_1')
        await asyncio.sleep(3 * period)  # <-- sleeping *longer* than the period, so that calls overlap
        return 'foo_1', call_time

    @shared_decorator
    async def foo_2() -> Tuple[str, float]:
        call_time = time.time()
        print('foo_2')
        await asyncio.sleep(3 * period)  # <-- sleeping *longer* than the period, so that calls overlap
        return 'foo_2', call_time

    def _assert_results(expected_name: str, res: List[Tuple[str, float]]) -> None:
        res = sorted(res)  # likely already sorted, based on how the even loop schedules tasks, but that's not guaranteed, so we sort here
        assert len(res) == 10

        for name, _ in res:
            assert name == expected_name

        for (_, a), (_, b) in zip(res, res[1:]):
            diff = b - a
            assert (0.5 * period) < diff < (1.5 * period)

    tasks = [
        f()
        for _ in range(10)
        for f in [foo_1, foo_2]
    ]

    res = await asyncio.gather(*tasks)
    assert len(res) == 20

    foo_1_res = res[0::2]
    foo_2_res = res[1::2]

    _assert_results('foo_1', foo_1_res)
    _assert_results('foo_2', foo_2_res)


@pytest.mark.asyncio
async def test_async_throttle_max_per_second__immediate_cases():
    @async_throttle(max_per_second=10)
    async def foo() -> float:
        call_time = time.time()
        return call_time

    # Ensure *first* call is immediate:
    t1 = time.time()
    t2 = await foo()
    assert 0.0 <= t2 - t1 < 0.05

    # Ensure *second* call is delayed:
    t1 = time.time()
    t2 = await foo()
    assert 0.05 <= t2 - t1 < 0.15

    # Ensure call is immediate if *enough time* has passed already:
    await asyncio.sleep(0.3)
    t1 = time.time()
    t2 = await foo()
    assert 0.0 <= t2 - t1 < 0.05

    # Ensure we delay *again*, if called without delay:
    t1 = time.time()
    t2 = await foo()
    assert 0.05 <= t2 - t1 < 0.15


def test_tool_param_passthrough():
    async def foo(a: int, b: str) -> float:
        """
        a func foo
        :param: a: int: the a param is an int
        :param: b: str: the b param is a str
        """
        return 3.14

    correct = (
        'a func foo',
        [
            {
                'name': 'a',
                'type': 'int',
                'description': 'the a param is an int',
            },
            {
                'name': 'b',
                'type': 'str',
                'description': 'the b param is a str',
            },
        ],
    )

    assert get_name(foo) == 'foo'
    assert get_tool_params(foo) == correct

    d1 = async_throttle(max_concurrent=2)
    d2 = async_throttle(max_per_second=10)
    d3 = async_throttle(max_concurrent=2, max_per_second=10)

    for d in [d1, d2, d3]:
        assert get_name(d(foo)) == 'foo'
        assert get_tool_params(d(foo)) == correct



from lasagna import (
    make_model_binder,
    known_models,
    flat_messages,
    human_input,
    to_str,
)
from lasagna import Model, EventCallback, AgentRun, Message
from lasagna.tui import tui_event_callback
from lasagna.planner.agent import build_default_planning_agent
from lasagna.planner.debug_model import DebugModel
from lasagna.async_util import async_throttle

import os
import sys
import json
import asyncio
import aiohttp
from datetime import datetime, timezone

from bs4 import BeautifulSoup

from typing import Dict, List

from dotenv import load_dotenv; load_dotenv()


if os.environ.get('DEBUG', 'false').lower() in ['1', 'true', 'yes', 'y']:
    print('Using DEBUG Model')
    BINDER = make_model_binder(DebugModel, 'NA')
elif os.environ.get('ANTHROPIC_API_KEY'):
    print('Using Anthropic')
    BINDER = known_models.anthropic_claude_haiku_4_5_binder
elif os.environ.get('OPENAI_API_KEY'):
    print('Using OpenAI')
    BINDER = known_models.openai_gpt_5_mini_binder
else:
    assert False, "Neither ANTHROPIC_API_KEY nor OPENAI_API_KEY is set! We need at least one to do this demo."


class fetch_url:
    """
    This tool fetches the supplied `url` (via an HTTP GET request).
    This tool returns the content the page at the supplied URL (`url`).
    :param: url: str: the URL to fetch
    """
    async def __call__(
        self,
        model: Model,
        event_callback: EventCallback,
        prev_runs: List[AgentRun],
    ) -> AgentRun:
        # These asserts are part of the layered agent contract.
        # See tools_util._run_single_tool()
        assert len(prev_runs) == 1
        only_run = prev_runs[0]
        assert only_run['agent'] == '__upstream__'
        assert only_run['type'] == 'messages'
        messages = only_run['messages']
        assert len(messages) > 0
        last_message = messages[-1]
        assert last_message['role'] == 'tool_call'
        tools = last_message['tools']
        assert len(tools) == 1
        tool = tools[0]

        # This tool's signature:
        args = json.loads(tool['function']['arguments'])
        assert 'url' in args
        url = args['url']
        assert isinstance(url, str)

        # Grab the URL's GET content:
        headers = {'User-Agent': os.environ.get('UA_STRING', 'Lasagna Planner Demo')}
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(url) as response:
                status = response.status
                content_type = response.content_type
                content = await response.text()

        # Use bs4 to strip away js and css (to reduce token count):
        if 'text/html' in content_type.lower():
            soup = BeautifulSoup(content)
            for el in soup(["script", "style"]):
                el.extract()
            content = soup.get_text()

        # Include the status code if it's not the expected 200:
        if status != 200:
            content = f"response HTTP status code: {status}\n\n{content}"

        # Prompt the downstream model:
        prompt_messages: List[Message] = [
            {
                'role': 'system',
                'text': 'Transcribe the important parts of this webpage into Markdown.',
            },
            {
                'role': 'human',
                'text': content,
            },
        ]
        new_messages = await model.run(
            event_callback,
            messages = prompt_messages,
            tools = [],
        )

        # Return the generated content:
        generated_messages = [*prompt_messages, *new_messages]
        return flat_messages('fetch_url', generated_messages)


@async_throttle(
    max_concurrent = int(os.environ.get('BRAVE_API_CNCR', 1)),
    max_per_second = 0.5 * float(os.environ.get('BRAVE_API_RPS', 1.0)),
)
async def web_search(search_term: str, num_results: int) -> str:
    """
    Use this tool to perform a web search to retrieve documents relevant to
    the supplied `search_term` string.
    This tool returns a JSON-encoded payload of the web search results.
    NOTE: The returned results will _not_ contain the full content of each
    retrieved document; rather, the returned results will contain small snippets
    from each document along with a URL to use to fetch the whole document.
    You can use the `fetch_url` to fetch those URLs, if you so choose.
    :param: search_term: str: the search term to send to the search engine
    :param: num_results: int: the number of results that should be returned by this search (must be a positive value, at most 100)
    """
    assert (0 < num_results <= 100), f"num_results cannot be {num_results}; it must be in the range (0, 100]"
    headers: Dict[str, str] = {
        'x-subscription-token': os.environ['BRAVE_API_KEY'],
    }
    url = 'https://api.search.brave.com/res/v1/news/search'  # TODO: just news?
    params: Dict[str, str] = {
        'q': search_term,
        'safesearch': 'moderate',
        'count': str(num_results),
        'spellcheck': 'false',
        'extra_snippets': 'true',
    }
    async with aiohttp.ClientSession(raise_for_status=True, headers=headers) as session:
        async with session.get(url, params=params) as response:
            response = await response.json()
    return json.dumps(response)


def ask_user(question: str) -> str:
    """
    Use this tool to ask the user a clarifying question.
    For example, if the user's prompt was unclear, you can
    clarify it with them. Or if the user's prompt is clear,
    but it's missing a key piece of information, you can ask
    them for that piece of information.
    This tool returns the user's answer to your question.
    :param: question: str: the question you need to ask the user
    """
    return input(f'AI needs to know: {question}\n>>> ')


def user_location() -> str:
    """
    Use this tool to get the user's current location.
    Use this tool if you need to know the user's location to answer
    local-specific questions.
    """
    return os.environ.get('USER_LOCATION', '[user location is unknown]')


def current_time() -> str:
    """
    Use this tool to obtain the _current_ date and time.
    This is important if you are asked to do something relative
    to the current time, then you can use this tool to get the
    current date and time. Otherwise, you don't need this tool.
    The date and time will be returned as a string in the user's
    current timezone.
    """
    utc_now = datetime.now(timezone.utc)
    user_tz_name = os.environ.get('USER_TIMEZONE', 'US/Central')
    if sys.version_info >= (3, 9):
        from zoneinfo import ZoneInfo
        user_tz = ZoneInfo(user_tz_name)
    else:
        import pytz
        user_tz = pytz.timezone(user_tz_name)
    user_now = utc_now.astimezone(user_tz)
    user_now_str = user_now.isoformat()
    return '\n'.join([
        f'The user is in the timezone: {user_tz_name}',
        f'Now is {user_now_str}',
    ])


async def run_planner(prompt: str) -> None:
    # TODO Tools:
    #   - access to file system (i.e. "agent memory")
    #   - access to a Python shell
    planning_agent = build_default_planning_agent(
        binder = BINDER,
        tools = [
            fetch_url(),
            web_search,
            ask_user,
            user_location,
            current_time,
        ],
    )
    run = await planning_agent(tui_event_callback, human_input(prompt))
    run_as_str = to_str(run)
    with open('planner_output.json', 'wt') as f:
        f.write(run_as_str)
    print(run_as_str)


def main() -> None:
    if sys.stdin.isatty():
        prompt = input('> ').strip()       # single-line input
    else:
        prompt = sys.stdin.read().strip()  # input until EOF
    asyncio.run(run_planner(prompt))


if __name__ == '__main__':
    main()



from lasagna import (     # <-- pip install -U lasagna-ai[openai,anthropic,bedrock]
    Model,
    EventCallback,
    AgentRun,
    recursive_extract_messages,
    flat_messages,
    known_models,
)

from lasagna.tui import tui_input_loop

from typing import List, Callable

import asyncio

import sympy as sp    # type: ignore

from dotenv import load_dotenv; load_dotenv()


def evaluate_math_expression(expression: str) -> float:
    """
    This tool evaluates a math expression and returns the result.
    Pass math expression as a string, for example:
     - "3 * 6 + 1"
     - "cos(2 * pi / 3) + log(8)"
     - "(4.5/2) + (6.3/1.2)"
     - ... etc
    :param: expression: str: the math expression to evaluate
    """
    expr = sp.sympify(expression)
    result = float(expr.evalf())
    return result


@known_models.openai_gpt_5_mini_binder
async def my_basic_agent(
    model: Model,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRun:
    messages = recursive_extract_messages(prev_runs, from_tools=False, from_extraction=False)
    tools: List[Callable] = [
        evaluate_math_expression,
    ]
    new_messages = await model.run(event_callback, messages, tools=tools)
    this_run = flat_messages('my_basic_agent', new_messages)
    return this_run


async def main() -> None:
    system_prompt = """You are my math assistant. Your name is Matheo."""
    await tui_input_loop(my_basic_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



import pytest

from lasagna.planner.default_prompts import (
    PlanOutput,
    subtask_input_generator,
    answer_input_prompt,
    append_tool_spec_to_prompt,
)

from lasagna.agent_util import extraction

from lasagna.types import AgentRun

from typing import List


PLAN_OBJ = PlanOutput(
    thoughts = '... hum ...',
    task_statement = 'do the thing',
    is_trivial = False,
    subtasks = ['first thing', 'next thing', 'last thing'],
)

EXTRACTION_RUN = extraction('test', [], PLAN_OBJ)


@pytest.mark.asyncio
async def test_subtask_input_generator():
    human_inputs: List[List[AgentRun]] = []
    async for human_input in subtask_input_generator(EXTRACTION_RUN):
        human_inputs.append(human_input)
    assert len(human_inputs) == 3
    for task_str, human_input in zip(PLAN_OBJ.subtasks, human_inputs):
        assert len(human_input) == 1
        single_input = human_input[0]
        assert single_input['type'] == 'messages'
        messages = single_input['messages']
        assert len(messages) == 1
        message = messages[0]
        assert message['role'] == 'human'
        assert message['text']
        assert task_str in message['text']


@pytest.mark.asyncio
async def test_answer_input_prompt():
    human_input = await answer_input_prompt(EXTRACTION_RUN)
    assert len(human_input) == 1
    single_input = human_input[0]
    assert single_input['type'] == 'messages'
    messages = single_input['messages']
    assert len(messages) == 1
    message = messages[0]
    assert message['role'] == 'human'
    assert message['text']
    assert PLAN_OBJ.task_statement in message['text']


def test_append_tool_spec_to_prompt():
    # No tools:
    new_prompt = append_tool_spec_to_prompt('do well', tools = [])
    assert new_prompt == 'do well'

    # With tools:
    def foo(a: int, b: int) -> float:
        """
        Divide b into a.
        :param: a: int: numerator
        :param: b: int: denominator
        """
        return a / b
    new_prompt = append_tool_spec_to_prompt('do well', tools = [
        foo,
    ])
    assert 'do well' in new_prompt
    assert 'foo' in new_prompt
    assert 'Divide b into a.' in new_prompt
    assert 'numerator' in new_prompt
    assert 'denominator' in new_prompt
    assert 'BLABLABLA' not in new_prompt



from lasagna import (     # <-- pip install -U lasagna-ai[openai,anthropic,bedrock]
    Model,
    EventCallback,
    AgentRun,
    recursive_extract_messages,
    flat_messages,
    known_models,
)

from lasagna.tui import tui_input_loop

from typing import List

import asyncio

from dotenv import load_dotenv; load_dotenv()


@known_models.openai_gpt_5_mini_binder
async def my_basic_agent(
    model: Model,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRun:
    messages = recursive_extract_messages(prev_runs, from_tools=False, from_extraction=False)
    new_messages = await model.run(event_callback, messages, tools=[])
    this_run = flat_messages('my_basic_agent', new_messages)
    return this_run


async def main() -> None:
    system_prompt = """You are a grumpy assistant. Be helpful, brief, and grumpy. Your name is Grumble."""
    await tui_input_loop(my_basic_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



from typing import List

from .types import ModelRecord

from .agent_util import make_model_binder


OPENAI_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'gpt-5-2025-08-07',
        'display_name': 'GPT-5',
    },
    {
        'formal_name': 'gpt-5-mini-2025-08-07',
        'display_name': 'GPT-5 mini',
    },
    {
        'formal_name': 'gpt-5-nano-2025-08-07',
        'display_name': 'GPT-5 nano',
    },
    {
        'formal_name': 'gpt-4.1-2025-04-14',
        'display_name': 'GPT-4.1',
    },
    {
        'formal_name': 'gpt-4o-mini-2024-07-18',
        'display_name': 'GPT-4o mini',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-4o-2024-11-20',
        'display_name': 'GPT-4o',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-4o-2024-08-06',
        'display_name': 'GPT-4o',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-4o-2024-05-13',
        'display_name': 'GPT-4o',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-4-turbo-2024-04-09',
        'display_name': 'GPT-4',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-3.5-turbo-0125',
        'display_name': 'GPT-3.5',
        'outdated': True,
    },
]

openai_gpt_5_binder        = make_model_binder('openai', 'gpt-5-2025-08-07')
openai_gpt_5_mini_binder   = make_model_binder('openai', 'gpt-5-mini-2025-08-07')
openai_gpt_5_nano_binder   = make_model_binder('openai', 'gpt-5-nano-2025-08-07')
openai_gpt_4_1_binder      = make_model_binder('openai', 'gpt-4.1-2025-04-14')


ANTHROPIC_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'claude-opus-4-1-20250805',
        'display_name': 'Claude Opus 4.1',
    },
    {
        'formal_name': 'claude-sonnet-4-5-20250929',
        'display_name': 'Claude Sonnet 4.5',
    },
    {
        'formal_name': 'claude-haiku-4-5-20251001',
        'display_name': 'Claude Haiku 4.5',
    },
    {
        'formal_name': 'claude-3-5-haiku-20241022',
        'display_name': 'Claude Haiku 3.5',
        'outdated': True,
    },
    {
        'formal_name': 'claude-opus-4-20250514',
        'display_name': 'Claude Opus 4',
        'outdated': True,
    },
    {
        'formal_name': 'claude-sonnet-4-20250514',
        'display_name': 'Claude Sonnet 4',
        'outdated': True,
    },
    {
        'formal_name': 'claude-3-7-sonnet-20250219',
        'display_name': 'Claude Sonnet 3.7',
        'outdated': True,
    },
    {
        'formal_name': 'claude-3-5-sonnet-20240620',
        'display_name': 'Claude Sonnet 3.5',
        'outdated': True,
    },
    {
        'formal_name': 'claude-3-opus-20240229',
        'display_name': 'Claude Opus 3',
        'outdated': True,
    },
    {
        'formal_name': 'claude-3-sonnet-20240229',
        'display_name': 'Claude Sonnet 3',
        'outdated': True,
    },
    {
        'formal_name': 'claude-3-haiku-20240307',
        'display_name': 'Claude Haiku 3',
        'outdated': True,
    },
]

anthropic_claude_opus_4_1_binder  = make_model_binder('anthropic', 'claude-opus-4-1-20250805')
anthropic_claude_sonnet_4_5_binder  = make_model_binder('anthropic', 'claude-sonnet-4-5-20250929')
anthropic_claude_haiku_4_5_binder = make_model_binder('anthropic', 'claude-haiku-4-5-20251001')


OLLAMA_KNOWN_MODELS: List[ModelRecord] = [
    # We'll only list models here that (1) support tool-calling because
    # that's kind of the whole point of lasagna, and (2) we've tested
    # ourselves and pass our "vibe" test. Users are of course more than
    # welcome to use *other* Ollama models as they see fit.
    {
        'formal_name': 'llama3.2',
        'display_name': 'Meta Llama 3.2',
    },
    {
        'formal_name': 'mistral-small',
        'display_name': 'Mistral Small',
    },
    {
        'formal_name': 'mistral-large',
        'display_name': 'Mistral Large',
    },
]

ollama_llama3_2_binder = make_model_binder('ollama', 'llama3.2')
ollama_mistral_small_binder = make_model_binder('ollama', 'mistral-small')
ollama_mistral_large_binder = make_model_binder('ollama', 'mistral-large')


BEDROCK_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'us.anthropic.claude-opus-4-1-20250805-v1:0',
        'display_name': 'Claude Opus 4.1',
    },
    {
        'formal_name': 'us.anthropic.claude-sonnet-4-20250514-v1:0',
        'display_name': 'Claude Sonnet 4',
    },
    {
        'formal_name': 'us.anthropic.claude-3-5-haiku-20241022-v1:0',
        'display_name': 'Claude Haiku 3.5',
    },
    {
        'formal_name': 'us.anthropic.claude-opus-4-20250514-v1:0',
        'display_name': 'Claude Opus 4',
        'outdated': True,
    },
    {
        'formal_name': 'us.anthropic.claude-3-7-sonnet-20250219-v1:0',
        'display_name': 'Claude Sonnet 3.7',
        'outdated': True,
    },
    {
        'formal_name': 'us.anthropic.claude-3-5-sonnet-20240620-v1:0',
        'display_name': 'Claude Sonnet 3.5',
        'outdated': True,
    },
    {
        'formal_name': 'us.anthropic.claude-3-5-sonnet-20241022-v2:0',
        'display_name': 'Claude Sonnet 3.5 (v2)',
        'outdated': True,
    },
    {
        'formal_name': 'us.anthropic.claude-3-opus-20240229-v1:0',
        'display_name': 'Claude Opus 3',
        'outdated': True,
    },
    {
        'formal_name': 'us.anthropic.claude-3-sonnet-20240229-v1:0',
        'display_name': 'Claude Sonnet 3',
        'outdated': True,
    },
    {
        'formal_name': 'us.anthropic.claude-3-haiku-20240307-v1:0',
        'display_name': 'Claude Haiku 3',
        'outdated': True,
    },
]

bedrock_claude_opus_4_1_binder = make_model_binder('bedrock', 'us.anthropic.claude-opus-4-1-20250805-v1:0')
bedrock_claude_sonnet_4_binder = make_model_binder('bedrock', 'us.anthropic.claude-sonnet-4-20250514-v1:0')
bedrock_claude_haiku_3_5_binder = make_model_binder('bedrock', 'us.anthropic.claude-3-5-haiku-20241022-v1:0')


abstract_binder = make_model_binder('__abstract__', '')


NVIDIA_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'meta/llama3-70b-instruct',
        'display_name': 'meta/llama3-70b-instruct',
    },
    {
        'formal_name': 'meta/llama3-8b-instruct',
        'display_name': 'meta/llama3-8b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.1-8b-instruct',
        'display_name': 'meta/llama-3.1-8b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.1-70b-instruct',
        'display_name': 'meta/llama-3.1-70b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.1-405b-instruct',
        'display_name': 'meta/llama-3.1-405b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-1b-instruct',
        'display_name': 'meta/llama-3.2-1b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-3b-instruct',
        'display_name': 'meta/llama-3.2-3b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-11b-vision-instruct',
        'display_name': 'meta/llama-3.2-11b-vision-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-90b-vision-instruct',
        'display_name': 'meta/llama-3.2-90b-vision-instruct',
    },
    {
        'formal_name': 'mistralai/mistral-large',
        'display_name': 'mistralai/mistral-large',
    },
    {
        'formal_name': 'mistralai/codestral-22b-instruct-v0.1',
        'display_name': 'mistralai/codestral-22b-instruct-v0.1',
    },
    {
        'formal_name': 'mistralai/mixtral-8x22b-instruct-v0.1',
        'display_name': 'mistralai/mixtral-8x22b-instruct-v0.1',
    },
    {
        'formal_name': 'mistralai/mixtral-8x7b-instruct-v0.1',
        'display_name': 'mistralai/mixtral-8x7b-instruct-v0.1',
    },
    {
        'formal_name': 'google/gemma-7b',
        'display_name': 'google/gemma-7b',
    },
    {
        'formal_name': 'google/recurrentgemma-2b',
        'display_name': 'google/recurrentgemma-2b',
    },
    {
        'formal_name': 'microsoft/phi-3-mini-128k-instruct',
        'display_name': 'microsoft/phi-3-mini-128k-instruct',
    },
]

nvidia_meta_llama3_70b_instruct_binder           = make_model_binder('nvidia', 'meta/llama3-70b-instruct')
nvidia_meta_llama3_8b_instruct_binder            = make_model_binder('nvidia', 'meta/llama3-8b-instruct')
nvidia_meta_llama3_1_8b_instruct_binder          = make_model_binder('nvidia', 'meta/llama-3.1-8b-instruct')
nvidia_meta_llama3_1_70b_instruct_binder         = make_model_binder('nvidia', 'meta/llama-3.1-70b-instruct')
nvidia_meta_llama3_1_405b_instruct_binder        = make_model_binder('nvidia', 'meta/llama-3.1-405b-instruct')
nvidia_meta_llama3_2_1b_instruct_binder          = make_model_binder('nvidia', 'meta/llama-3.2-1b-instruct')
nvidia_meta_llama3_2_3b_instruct_binder          = make_model_binder('nvidia', 'meta/llama-3.2-3b-instruct')
nvidia_meta_llama3_2_11b_vision_instruct_binder  = make_model_binder('nvidia', 'meta/llama-3.2-11b-vision-instruct')
nvidia_meta_llama3_2_90b_vision_instruct_binder  = make_model_binder('nvidia', 'meta/llama-3.2-90b-vision-instruct')

nvidia_mistralai_mistral_large_binder            = make_model_binder('nvidia', 'mistralai/mistral-large')
nvidia_mistralai_codestral_22b_instruct_binder   = make_model_binder('nvidia', 'mistralai/codestral-22b-instruct-v0.1')
nvidia_mistralai_mixtral_8x22b_instruct_binder   = make_model_binder('nvidia', 'mistralai/mixtral-8x22b-instruct-v0.1')
nvidia_mistralai_mixtral_8x7b_instruct_binder    = make_model_binder('nvidia', 'mistralai/mixtral-8x7b-instruct-v0.1')

nvidia_google_gemma_7b_binder                    = make_model_binder('nvidia', 'google/gemma-7b')
nvidia_google_recurrentgemma_2b_binder           = make_model_binder('nvidia', 'google/recurrentgemma-2b')

nvidia_microsoft_phi_3_mini_128k_instruct_binder = make_model_binder('nvidia', 'microsoft/phi-3-mini-128k-instruct')



from .agent_runner import run

from .util import get_name

from .pydantic_util import result_to_string

from .types import (
    AgentRunChained,
    AgentRunExtraction,
    AgentRunMessageList,
    AgentRunParallel,
    AgentSpec,
    AgentRun,
    Cost,
    ModelFactory,
    ModelRecord,
    AgentCallable,
    BoundAgentCallable,
    EventCallback,
    EventPayload,
    Message,
    Model,
    ExtractionType,
    ToolResult,
)

from typing import (
    Union, List, Any, Type,
    Callable, Protocol,
)

import pydantic
import json
import copy


def make_model_binder(
    provider: Union[str, ModelFactory],
    model: Union[str, ModelRecord],
    **model_kwargs: Any,
) -> Callable[[AgentCallable], BoundAgentCallable]:
    class ModelBinder():
        def __call__(self, agent: AgentCallable) -> BoundAgentCallable:
            spec: AgentSpec = {
                'agent': agent,
                'provider': provider,
                'model': model,
                'model_kwargs': model_kwargs,
            }
            async def bound_agent(event_callback: EventCallback, prev_runs: List[AgentRun]) -> AgentRun:
                return await run(spec, event_callback, prev_runs)
            for attr in ['__module__', '__qualname__', '__doc__']:
                if hasattr(agent, attr):
                    setattr(bound_agent, attr, getattr(agent, attr))
            bound_agent.__name__ = get_name(agent)
            return bound_agent

        def __str__(self) -> str:
            info = (provider, model, model_kwargs)
            return f'model binder: {info}'

    return ModelBinder()


class PartiallyBoundAgentCallable(Protocol):
    def __call__(
        self,
        **model_kwargs: Any,
    ) -> Callable[[AgentCallable], BoundAgentCallable]: ...


def make_partial_model_binder(
    provider: Union[str, ModelFactory],
    model: Union[str, ModelRecord],
) -> PartiallyBoundAgentCallable:
    class PartialModelBinder():
        def __call__(
            self,
            **model_kwargs: Any,
        ) -> Callable[[AgentCallable], BoundAgentCallable]:
            return make_model_binder(provider, model, **model_kwargs)

        def __str__(self) -> str:
            info = (provider, model)
            return f'partial model binder: {info}'

    return PartialModelBinder()


def _extract_messages_from_tool_result(
    tool_result: ToolResult,
    from_extraction: bool,
) -> List[Message]:
    if tool_result['type'] == 'layered_agent':
        run = tool_result['run']
        return recursive_extract_messages(run, from_tools=True, from_extraction=from_extraction)
    return []


def _recursive_extract_messages_from_tool_res(
    messages: List[Message],
    from_extraction: bool,
) -> List[Message]:
    ms: List[Message] = []
    for m in messages:
        ms.append(m)
        if m['role'] == 'tool_res':
            for t in m['tools']:
                ms.extend(_extract_messages_from_tool_result(t, from_extraction=from_extraction))
    return ms


def recursive_extract_messages(
    agent_run_or_runs: Union[AgentRun, List[AgentRun]],
    from_tools: bool,
    from_extraction: bool,
) -> List[Message]:
    """DFS retrieve all messages within a list of `AgentRuns`."""
    agent_runs: List[AgentRun]
    if isinstance(agent_run_or_runs, list):
        agent_runs = agent_run_or_runs
    else:
        agent_runs = [agent_run_or_runs]
    messages: List[Message] = []
    for run in agent_runs:
        if run['type'] == 'messages':
            messages.extend(
                (
                    _recursive_extract_messages_from_tool_res(run['messages'], from_extraction=from_extraction)
                    if from_tools else
                    run['messages']
                ),
            )
        elif run['type'] == 'chain' or run['type'] == 'parallel':
            messages.extend(
                recursive_extract_messages(run['runs'], from_tools=from_tools, from_extraction=from_extraction),
            )
        elif run['type'] == 'extraction':
            if from_extraction:
                messages.extend(
                    (
                        _recursive_extract_messages_from_tool_res(run['messages'], from_extraction=from_extraction)
                        if from_tools else
                        run['messages']
                    ),
                )
            else:
                messages.append({
                    'role': 'ai',
                    'text': result_to_string(run['result']),
                })
        else:
            raise RuntimeError('unreachable')
    return messages


def extract_last_message(
    agent_run_or_runs: Union[AgentRun, List[AgentRun]],
    from_tools: bool,
    from_extraction: bool,
) -> Message:
    messages = recursive_extract_messages(agent_run_or_runs, from_tools=from_tools, from_extraction=from_extraction)
    if len(messages) == 0:
        raise ValueError('no messages found')
    return messages[-1]


def flat_messages(
    agent_name: str,
    messages: List[Message],
) -> AgentRunMessageList:
    return {
        'agent': agent_name,
        'type': 'messages',
        'messages': messages,
    }


def parallel_runs(
    agent_name: str,
    runs: List[AgentRun],
) -> AgentRunParallel:
    return {
        'agent': agent_name,
        'type': 'parallel',
        'runs': runs,
    }


def chained_runs(
    agent_name: str,
    runs: List[AgentRun],
) -> AgentRunChained:
    return {
        'agent': agent_name,
        'type': 'chain',
        'runs': runs,
    }


def extraction(
    agent_name: str,
    messages: List[Message],
    result: Any,
) -> AgentRunExtraction:
    return {
        'agent': agent_name,
        'type': 'extraction',
        'messages': messages,
        'result': result,
    }


def human_input(prompt: str) -> List[AgentRun]:
    return [
        flat_messages('human_input', [
            {
                'role': 'human',
                'text': prompt,
            },
        ]),
    ]


def override_system_prompt(
    messages: List[Message],
    system_prompt: str,
) -> List[Message]:
    sp: Message = {
        'role': 'system',
        'text': system_prompt,
    }
    if messages and messages[0]['role'] == 'system':
        return [sp, *messages[1:]]
    else:
        return [sp, *messages]


def strip_tool_calls_and_results(
    messages: List[Message],
) -> List[Message]:
    return [
        m
        for m in messages
        if m['role'] != 'tool_call' and m['role'] != 'tool_res'
    ]


def strip_all_but_last_human_message(
    messages: List[Message],
) -> List[Message]:
    for m in reversed(messages):
        if m['role'] == 'human':
            return [m]
    return []


def recursive_sum_costs(
    agent_run_or_runs: Union[AgentRun, List[AgentRun]],
) -> Cost:
    messages = recursive_extract_messages(agent_run_or_runs, from_tools=True, from_extraction=True)

    input_tokens  = [m['cost']['input_tokens']  for m in messages if 'cost' in m and 'input_tokens'  in m['cost']]
    output_tokens = [m['cost']['output_tokens'] for m in messages if 'cost' in m and 'output_tokens' in m['cost']]
    total_tokens  = [m['cost']['total_tokens']  for m in messages if 'cost' in m and 'total_tokens'  in m['cost']]

    cost: Cost = {}

    if input_tokens:
        cost['input_tokens'] = sum(input_tokens)
    if output_tokens:
        cost['output_tokens'] = sum(output_tokens)
    if total_tokens:
        cost['total_tokens'] = sum(total_tokens)

    return cost


async def noop_callback(event: EventPayload) -> None:
    # "noop" mean "no operation" means DON'T DO ANYTHING!
    assert event


MessageExtractor = Callable[[List[AgentRun]], List[Message]]


def build_standard_message_extractor(
    extract_from_tools: bool = False,
    extract_from_extraction: bool = False,
    keep_only_last_human_message: bool = False,
    strip_tool_messages: bool = False,
    system_prompt_override: Union[str, None] = None,
) -> MessageExtractor:
    def extractor(prev_runs: List[AgentRun]) -> List[Message]:
        messages = recursive_extract_messages(
            prev_runs,
            from_tools = extract_from_tools,
            from_extraction = extract_from_extraction,
        )
        if keep_only_last_human_message:
            messages = strip_all_but_last_human_message(messages)
        if strip_tool_messages:
            messages = strip_tool_calls_and_results(messages)
        if system_prompt_override:
            messages = override_system_prompt(messages, system_prompt_override)
        return messages

    return extractor


default_message_extractor = build_standard_message_extractor()


def build_simple_agent(
    name: str,
    tools: List[Callable] = [],
    force_tool: bool = False,
    max_tool_iters: int = 5,
    message_extractor: MessageExtractor = default_message_extractor,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class SimpleAgent():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            messages = message_extractor(prev_runs)
            new_messages = await model.run(
                event_callback = event_callback,
                messages = messages,
                tools = tools,
                force_tool = force_tool,
                max_tool_iters = max_tool_iters,
            )
            return flat_messages(name, new_messages)

        def __str__(self) -> str:
            return name

    a = SimpleAgent()
    if doc:
        a.__doc__ = doc
    return a


def build_extraction_agent(
    name: str,
    extraction_type: Type[ExtractionType],
    message_extractor: MessageExtractor = default_message_extractor,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class ExtractionAgent():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            messages = message_extractor(prev_runs)
            message, result = await model.extract(event_callback, messages, extraction_type)
            return {
                'agent': name,
                'type': 'extraction',
                'messages': [message],
                'result': result,
            }

        def __str__(self) -> str:
            return name

    a = ExtractionAgent()
    if doc:
        a.__doc__ = doc
    return a


def build_agent_chainer(
    name: str,
    agents: List[BoundAgentCallable],
    message_extractor: Union[MessageExtractor, None] = None,
    doc: Union[str, None] = None,
) -> BoundAgentCallable:
    class ChainedAgents():
        async def __call__(
            self,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            if message_extractor is not None:
                prev_runs = [
                    flat_messages(
                        agent_name = name,
                        messages = message_extractor(prev_runs),
                    ),
                ]
            else:
                prev_runs = [*prev_runs]  # shallow copy
            new_runs: List[AgentRun] = []
            for agent in agents:
                this_run = await agent(event_callback, prev_runs)
                prev_runs.append(this_run)
                new_runs.append(this_run)
            return {
                'agent': name,
                'type': 'chain',
                'runs': new_runs,
            }

        def __str__(self) -> str:
            return name

    a = ChainedAgents()
    if doc:
        a.__doc__ = doc
    return a


def build_agent_router(
    name: str,
    extraction_type: Type[ExtractionType],
    pick_agent_func: Callable[[ExtractionType], BoundAgentCallable],
    message_extractor: MessageExtractor = default_message_extractor,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class AgentRouter():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            messages = message_extractor(prev_runs)
            message, result = await model.extract(event_callback, messages, extraction_type)
            extraction: AgentRun = {
                'agent': name,
                'type': 'extraction',
                'messages': [message],
                'result': result,
            }
            agent = pick_agent_func(result)
            run = await agent(event_callback, prev_runs)
            return {
                'agent': name,
                'type': 'chain',
                'runs': [
                    extraction,
                    run,
                ],
            }

        def __str__(self) -> str:
            return name

    a = AgentRouter()
    if doc:
        a.__doc__ = doc
    return a


def build_static_output_agent(
    name: str,
    output: str,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class StaticOutputAgent():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            assert model is not None and prev_runs is not None
            await event_callback(('ai', 'text_event', output))
            return flat_messages(name, [
                {
                    'role': 'ai',
                    'text': output,
                },
            ])

        def __str__(self) -> str:
            return name

    a = StaticOutputAgent()
    if doc:
        a.__doc__ = doc
    return a


def _strip_raw_cost_from_toolres(toolres: ToolResult) -> ToolResult:
    if toolres['type'] == 'layered_agent':
        toolres = copy.copy(toolres)
        toolres['run'] = strip_raw_cost_from_run(toolres['run'])
        return toolres
    elif toolres['type'] == 'any':
        return toolres
    else:
        raise RuntimeError('unreachable')


def strip_raw_cost_from_message(message: Message) -> Message:
    message = copy.copy(message)
    if 'raw' in message:
        del message['raw']
    if 'cost' in message:
        del message['cost']
    if message['role'] == 'tool_res':
        message['tools'] = [
            _strip_raw_cost_from_toolres(t)
            for t in message['tools']
        ]
    return message


def strip_raw_cost_from_run(run: AgentRun) -> AgentRun:
    run = copy.copy(run)
    if run['type'] == 'messages' or run['type'] == 'extraction':
        run['messages'] = [strip_raw_cost_from_message(m) for m in run['messages']]
    elif run['type'] == 'chain' or run['type'] == 'parallel':
        run['runs'] = [strip_raw_cost_from_run(r) for r in run['runs']]
    else:
        raise RuntimeError('unreachable')
    return run


def _model_dump_pydantic_results_for_toolres(toolres: ToolResult) -> ToolResult:
    toolres = copy.copy(toolres)
    if toolres['type'] == 'layered_agent':
        toolres['run'] = model_dump_all_pydantic_results(toolres['run'])
    elif toolres['type'] == 'any':
        result = toolres['result']
        if isinstance(result, pydantic.BaseModel):
            toolres['result'] = result.model_dump()
    else:
        raise RuntimeError('unreachable')
    return toolres


def _model_dump_pydantic_results_for_message(message: Message) -> Message:
    if message['role'] == 'tool_res':
        message = copy.copy(message)
        message['tools'] = [
            _model_dump_pydantic_results_for_toolres(t)
            for t in message['tools']
        ]
    return message


def model_dump_all_pydantic_results(run: AgentRun) -> AgentRun:
    run = copy.copy(run)
    if run['type'] == 'messages':
        run['messages'] = [
            _model_dump_pydantic_results_for_message(m)
            for m in run['messages']
        ]
    elif run['type'] == 'extraction':
        run['messages'] = [
            _model_dump_pydantic_results_for_message(m)
            for m in run['messages']
        ]
        result = run['result']
        if isinstance(result, pydantic.BaseModel):
            run['result'] = result.model_dump()
    elif run['type'] == 'chain' or run['type'] == 'parallel':
        run['runs'] = [model_dump_all_pydantic_results(r) for r in run['runs']]
    else:
        raise RuntimeError('unreachable')
    return run


def _prep_for_debug_print(run: AgentRun) -> AgentRun:
    run = model_dump_all_pydantic_results(run)
    run = strip_raw_cost_from_run(run)
    return run


def to_str(
    agent_run_or_runs: Union[AgentRun, List[AgentRun]],
) -> str:
    if isinstance(agent_run_or_runs, list):
        runs = agent_run_or_runs
        runs = [_prep_for_debug_print(r) for r in runs]
        return json.dumps(runs, indent=2)
    else:
        run = _prep_for_debug_print(agent_run_or_runs)
        return json.dumps(run, indent=2)



from typing import (
    Callable,
    List,
    Union,
    Type,
    Awaitable,
    Optional,
)

from . import (
    build_simple_agent,
    build_extraction_agent,
    AgentCallable,
    BoundAgentCallable,
    EventCallback,
    EventPayload,
    AgentRun,
    flat_messages,
    noop_callback,
    Message,
)
from .tools_util import (
    extract_tool_result_as_sting,
)
from .types import ExtractionType


def _build_text_event_callback(simple_text_callback: Callable[[str], Awaitable[None]]) -> EventCallback:
    async def bound_simple_text_callback(event: EventPayload) -> None:
        if event[0] == 'ai' and event[1] == 'text_event':
            await simple_text_callback(event[2])
    return bound_simple_text_callback


def _extract_message_text(message: Message) -> str:
    if message["role"] == "system" or message["role"] == "human":
        return message["text"] or ""
    elif message["role"] == "ai":
        return message["text"] or ""
    elif message["role"] == "tool_res":
        return extract_tool_result_as_sting(message["tools"][-1])
    else:
        raise RuntimeError(f"unexpected message role: {message['role']}")


async def easy_ask(
    binder: Callable[[AgentCallable], BoundAgentCallable],
    prompt: str,
    system_prompt: Optional[str] = None,
    streaming_callback: Union[Callable[[str], Awaitable[None]], None] = None,
    tools: List[Callable] = [],
    force_tool: bool = False,
    max_tool_iters: int = 5,
) -> str:
    agent = binder(
        build_simple_agent(
            name = "easy_ask",
            tools = tools,
            force_tool = force_tool,
            max_tool_iters = max_tool_iters,
        )
    )

    messages: List[Message] = []
    if system_prompt:
        messages.append({
            'role': 'system',
            'text': system_prompt,
        })
    messages.append({
        'role': 'human',
        'text': prompt,
    })
    runs: List[AgentRun] = [
        flat_messages(
            agent_name = "easy_ask",
            messages = messages,
        )
    ]
    response: AgentRun = await agent(
        _build_text_event_callback(streaming_callback) if streaming_callback else noop_callback,
        runs,
    )
    assert response['type'] == 'messages'
    response_text: List[str] = []
    for message in response["messages"]:
        response_text.append(_extract_message_text(message))
    return "\n".join(response_text)


async def easy_extract(
    binder: Callable[[AgentCallable], BoundAgentCallable],
    prompt: str,
    extraction_type: Type[ExtractionType],
    system_prompt: Optional[str] = None,
    streaming_callback: Union[Callable[[str], Awaitable[None]], None] = None,
) -> ExtractionType:
    agent = binder(
        build_extraction_agent(
            name = "easy_extract",
            extraction_type = extraction_type,
        )
    )

    messages: List[Message] = []
    if system_prompt:
        messages.append({
            'role': 'system',
            'text': system_prompt,
        })
    messages.append({
        'role': 'human',
        'text': prompt,
    })

    response: AgentRun = await agent(
        _build_text_event_callback(streaming_callback) if streaming_callback else noop_callback,
        [
            flat_messages(
                agent_name = 'easy_extract',
                messages = messages,
            ),
        ],
    )

    assert response['type'] == 'extraction'
    assert isinstance(response['result'], extraction_type)
    return response['result']



from ..types import Model, EventCallback, Message, ExtractionType

from pydantic import BaseModel

import random
from typing import List, Tuple, Callable, Any, Type, cast


class DebugModel(Model):
    def __init__(self, model: str, **model_kwargs: Any) -> None:
        seed = model_kwargs.get('seed', 0)
        self.rand = random.Random()
        self.rand.seed(seed, version=2)

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        assert len(tools) == 0
        assert not force_tool
        return [{
            'role': 'ai',
            'text': f'The answer is: {len(messages)}',
        }]

    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        message: Message = {
            'role': 'ai',
            'text': f'Will extract from {len(messages)} messages...',
        }
        is_trivial = self.rand.choice([True, False])
        if len(messages) > 3:
            is_trivial = True  # don't go too deep!
        last_message = messages[-1]
        assert last_message['role'] == 'human'
        data = dict(
            thoughts = '',
            task_statement = last_message['text'] or '!!! NO TEXT !!!',
            is_trivial = is_trivial,
            subtasks = [] if is_trivial else self._rand_tasks(),
        )
        if issubclass(extraction_type, BaseModel):
            output = cast(ExtractionType, extraction_type.model_validate(data))
        else:
            raise ValueError(f'Cannot handle data for type: {extraction_type}')
        return message, output

    def _rand_tasks(self) -> List[str]:
        length = self.rand.randint(2, 5)  # inclusive!
        return [
            f'step #{i + 1}: bla bla bla'
            for i in range(length)
        ]

    def _text(self, messages: List[Message]) -> str:
        return ' || '.join(f"[[ {m['role']}: {m.get('text')} ]]" for m in messages)

    def config_hash(self) -> str:
        raise NotImplementedError('not needed and not implemented')



"""
The Lasagna Recursive Planning Agent (LRPA)

This planning agent ("planner") recursively breaks down a task until
each subtask (or subsubtask, or subsubsubtask, etc, as deep as needed)
becomes "trivial" (that is the recursion base case). It then aggregates
answers back up the recursion tree using the "answer agent" at each level
to build up higher-level answers until the final (original) task is
complete. It performs a DFS traversal, injecting answers to subtasks
into the history to inform future subtasks (at the same level) and
higher-level subtasks (when unwinding).

It uses the Lasagna "layered agent" ideas (both the hierarchical AgentRun
data structure as well as the agents-calling-agents idea). Recursive
planning agents were a core idea I had when I began this library, which
is why its design is the way it is, so it's great to get this LRPA
implementation started!

I still have some ideas left to implement, like:

1. Backtracking: What if the task list turns out to be bad? Sometimes
   you don't find that out until you are half-way through the task.
   A backtracking feature would allow the agent to bail on a plan
   and start over with a new plan (using what it learned to inform
   the new plan).

2. Loopback: Unlike backtracking, loopback keeps the _same_ plan but
   just returns back to a previous step that it found out needs revision.
   It then does the revision, and recomputes the later steps with those
   new/fixed results.

3. Retries: Sometimes you just need to try again. Same plan, same approach,
   just give it another try.

4. History Exclusion: Some subtasks don't need the full history, thus those
   subtasks should be launched _without_ the full history. This will help
   in three ways: (1) fewer tokens to pay for, (2) faster responses,
   (3) better answers (it's good not to include context when it's not needed).
   The planner will identify subtasks that can stand on their own,
   and kick those off without the full history. Another benefit: We can
   execute those in parallel! A more _generalized_ idea here is to generate
   subtask dependencies, which would enable optimal parallelization and
   precise history inclusion.

5. Watchdogs: After each generation (be it a subtask list, an answer, etc)
   we could have watchdog models judge those output and accept/reject them.
   Accepted outputs would move through the process as normal. Rejected outputs
   would be regenerated with injected feedback from the watchdog.

6. Timebound: Part of the generation should be governed by how much _time_
   is left to complete the original task. As time approaches the deadline,
   the generation should do less diving and more unwinding. A deadline is
   optional, but it may be helpful for open-ended tasks that could go on
   forever. Similar a John Cleese's point - creativity comes from reveling
   in the unknown for as long as you can, but also meeting deadlines.
   (PS: This feature may be implemented as a watchdog.)

7. Pins: "Put a pin in that." Since the planner is DFS, there's a danger
   in diving forever into the first subtask. The planner should be able
   to "pin" something and come back to it later (if times allows), so that
   it can begin unwinding a task and move on.

8. Standard Tooling:
   - web search,
   - memory (maybe via a file-system API),
      - namespaced to the task and its parents
   - ability to execute Python code,
   - etc.
"""

from ..types import (
    Model,
    EventCallback,
    AgentRun,
    Message,
    AgentCallable,
    BoundAgentCallable,
)
from ..agent_util import (
    override_system_prompt,
    build_extraction_agent,
    build_simple_agent,
    MessageExtractor,
)
from ..agent_util import chained_runs
from ..pydantic_util import result_to_string
from ..known_models import abstract_binder

from .util import extract_is_trivial
from . import default_prompts

import copy
from typing import (
    List, Tuple, Optional,
    Callable, AsyncIterable, Awaitable,
)


def build_default_planning_agent(
    *,
    binder: Callable[[AgentCallable], BoundAgentCallable],
    tools: List[Callable] = [],
    max_tool_iters: int = 5,
    max_depth: int = 4,
) -> BoundAgentCallable:
    return build_planning_agent(
        extraction_agent = binder(build_extraction_agent(
            'extraction_agent',
            extraction_type = default_prompts.PlanOutput,
            message_extractor = build_message_extractor_lrpa_aware(
                system_prompt_override = default_prompts.append_tool_spec_to_prompt(
                    default_prompts.PLANNING_EXTRACTION_AGENT_SYSTEM_PROMPT,
                    tools,
                ),
            ),
        )),
        answer_agent = binder(build_simple_agent(
            'answer_agent',
            tools = tools,
            max_tool_iters = max_tool_iters,
            message_extractor = build_message_extractor_lrpa_aware(
                system_prompt_override = None,
            ),
        )),
        subtask_input_generator = default_prompts.subtask_input_generator,
        answer_input_prompt = default_prompts.answer_input_prompt,
        max_depth = max_depth,
    )


def build_planning_agent(
    *,
    extraction_agent: BoundAgentCallable,
    answer_agent: BoundAgentCallable,
    subtask_input_generator: Callable[[AgentRun], AsyncIterable[List[AgentRun]]],
    answer_input_prompt: Callable[[AgentRun], Awaitable[List[AgentRun]]],
    max_depth: int,
) -> BoundAgentCallable:
    @abstract_binder
    async def planning_agent(
        _: Model,
        event_callback: EventCallback,
        prev_runs: List[AgentRun],
    ) -> AgentRun:
        # We want to be **pure**, but also not copy more than we need to.
        # So, `prev_runs` will be minimally copied, and `planning_agent_chain`
        # will be inserted into the copied `prev_runs` in the correct location.
        # That is: Modifying `planning_agent_chain` will modify the *copied*
        #          `prev_runs`, which is fine since it's a copy. The code below
        #          will append its output into `planning_agent_chain`.
        prev_runs, planning_agent_chain, depth = _minimal_copy_lrpa_aware(prev_runs)

        # All Lasagna agents return a *new* AgentRun representing
        # the new content generated by this agent. We'll keep track
        # of all the content generated by this agent in the `output_runs`
        # list below.
        assert planning_agent_chain['type'] == 'chain'
        output_runs = planning_agent_chain['runs']

        extraction_run = await extraction_agent(
            event_callback,
            prev_runs,
        )
        output_runs.append(extraction_run)

        is_trivial = extract_is_trivial(extraction_run)

        if not is_trivial and depth < max_depth:
            subtask_chain_of_chains = chained_runs('subtask_chain_of_chains', [])
            output_runs.append(subtask_chain_of_chains)
            async for subtask_input in subtask_input_generator(extraction_run):
                subtask_chain = chained_runs('subtask_chain', [
                    *subtask_input,
                ])
                subtask_chain_of_chains['runs'].append(subtask_chain)
                subtask_run = await planning_agent(
                    event_callback,
                    prev_runs,
                )
                subtask_chain['runs'].append(subtask_run)

        answer_chain = chained_runs('answer_chain', [
            chained_runs('answer_input_chain', [
                *(await answer_input_prompt(extraction_run)),
            ]),
        ])
        output_runs.append(answer_chain)
        answer_run = await answer_agent(
            event_callback,
            prev_runs,
        )
        answer_chain['runs'].append(answer_run)

        return planning_agent_chain

    return planning_agent


_PLANNER_RECURSION_AGENTS = ['planning_agent_chain', 'subtask_chain_of_chains', 'subtask_chain']


def _minimal_copy_lrpa_aware(
    prev_runs: List[AgentRun],
    depth: int = 0,
) -> Tuple[List[AgentRun], AgentRun, int]:
    """
    Minimally copy `prev_runs` and add a 'planning_agent_chain'
    to the end of it at the deepest depth. This implementation
    assumes the structure created by `planning_agent`, where the
    new output always goes at the end of the deepest planning chain.
    """
    if not prev_runs:
        raise ValueError("empty `prev_runs`")

    last_run = prev_runs[-1]

    if last_run['agent'] in _PLANNER_RECURSION_AGENTS:
        assert last_run['type'] == 'chain'
        last_run_copy = copy.copy(last_run)     # shallow copy
        (
            last_run_copy['runs'],
            deepest_planning_agent_chain,
            deepest_depth,
        ) = _minimal_copy_lrpa_aware(
            last_run_copy['runs'],
            depth + 1 if last_run['agent'] == 'planning_agent_chain' else depth
        )
        prev_runs_copy = copy.copy(prev_runs)   # shallow copy
        prev_runs_copy[-1] = last_run_copy
        return prev_runs_copy, deepest_planning_agent_chain, deepest_depth

    # This is the _first_ call to the planning agent (no recursion yet).
    # Or it's the _deepest_ call. Either way, we're at the base case.
    # So merely append the root of a new planning agent output to `prev_runs`.
    planning_agent_chain = chained_runs('planning_agent_chain', [])
    prev_runs_copy = copy.copy(prev_runs)  # shallow copy
    prev_runs_copy.append(planning_agent_chain)
    return prev_runs_copy, planning_agent_chain, depth


def build_message_extractor_lrpa_aware(
    system_prompt_override: Optional[str],
) -> MessageExtractor:
    def extract_messages_lrpa_aware(
        prev_runs: List[AgentRun],
        depth: int = 0,
    ) -> List[Message]:
        """
        This function extracts only _relevant_ messages. If a subtask has an
        answer, the subtasks below it are removed (since the answer is what
        matters in this case). If a subtask does not yet have an answer, its
        full surrounding context is preserved so that an answer can be reached.
        Stated another way: The subtask tree is pruned to keep only the messages
        relevant to obtain the next answer.
        NOTE: The implementation of this function is tied to and assumes this
        is called from the `planning_agent`. That is, it assumes the hierarchical/
        recursive `prev_runs` in the input match that built by `planning_agent`.
        """
        messages: List[Message] = []

        for run in prev_runs:
            if run['agent'] == 'planning_agent_chain':
                assert run['type'] == 'chain'
                pa_runs = run['runs']
                if len(pa_runs) > 0 and pa_runs[-1]['agent'] == 'answer_chain':
                    extraction_run = pa_runs[0]
                    answer_chain = pa_runs[-1]
                    assert answer_chain['type'] == 'chain'
                    if len(answer_chain['runs']) > 1:
                        # We *must* have an answer... since it goes in the second spot of this answer_chain.
                        # Thus, we hit a special case. We'll *exclude* the intermediate steps, and just include this answer.
                        messages.extend(
                            extract_messages_lrpa_aware([extraction_run, answer_chain], depth = depth + 1),
                        )
                        continue

            if run['type'] == 'messages':
                messages.extend(
                    run['messages'],
                )

            elif run['type'] == 'chain' or run['type'] == 'parallel':
                messages.extend(
                    extract_messages_lrpa_aware(run['runs'], depth = depth + 1),
                )

            elif run['type'] == 'extraction':
                messages.append({
                    'role': 'ai',
                    'text': result_to_string(run['result']),
                })

            else:
                raise RuntimeError(f"unknown type: {run['type']}")

        if depth == 0 and system_prompt_override is not None:
            messages = override_system_prompt(messages, system_prompt_override)

        #if depth == 0:
        #    from ..agent_util import to_str
        #    print(to_str(prev_runs))
        #    import json
        #    print(json.dumps(messages, indent=2))
        #    print('-' * 60)

        return messages

    return extract_messages_lrpa_aware



import pytest

from lasagna.abstract_model import AbstractModel

from lasagna.agent_util import noop_callback


@pytest.mark.asyncio
async def test_abstract_model():
    model = AbstractModel(model = '')

    with pytest.raises(NotImplementedError):
        await model.run(
            event_callback = noop_callback,
            messages = [],
            tools = [],
        )

    with pytest.raises(NotImplementedError):
        await model.extract(
            event_callback = noop_callback,
            messages = [],
            extraction_type = int,
        )



from .types import *

from .agent_runner import run

from .agent_util import (
    make_model_binder,
    make_partial_model_binder,
    recursive_extract_messages,
    extract_last_message,
    flat_messages,
    parallel_runs,
    chained_runs,
    extraction,
    human_input,
    override_system_prompt,
    strip_tool_calls_and_results,
    strip_all_but_last_human_message,
    recursive_sum_costs,
    noop_callback,
    MessageExtractor,
    build_standard_message_extractor,
    default_message_extractor,
    build_simple_agent,
    build_extraction_agent,
    build_agent_chainer,
    build_agent_router,
    build_static_output_agent,
    strip_raw_cost_from_message,
    strip_raw_cost_from_run,
    model_dump_all_pydantic_results,
    to_str,
)

from .pydantic_util import (
    result_to_string,
)

from .easy import (
    easy_ask,
    easy_extract,
)

__all__ = [
    'run',
    'make_model_binder',
    'make_partial_model_binder',
    'recursive_extract_messages',
    'extract_last_message',
    'flat_messages',
    'parallel_runs',
    'chained_runs',
    'extraction',
    'human_input',
    'override_system_prompt',
    'strip_tool_calls_and_results',
    'strip_all_but_last_human_message',
    'recursive_sum_costs',
    'noop_callback',
    'MessageExtractor',
    'build_standard_message_extractor',
    'default_message_extractor',
    'build_simple_agent',
    'build_extraction_agent',
    'build_agent_chainer',
    'build_agent_router',
    'build_static_output_agent',
    'strip_raw_cost_from_message',
    'strip_raw_cost_from_run',
    'model_dump_all_pydantic_results',
    'to_str',
    'result_to_string',
    'easy_ask',
    'easy_extract',
]

__version__ = "0.20.0"



from lasagna.types import (
    EventCallback,
    Message,
    Model,
    ExtractionType,
)

from lasagna.util import recursive_hash

from typing import List, Tuple, Any, Callable, Type


class AbstractModel(Model):
    def __init__(self, model: str, **model_kwargs: Any):
        assert not model
        assert not model_kwargs

    def config_hash(self) -> str:
        return recursive_hash(None, {
            'provider': '__abstract__',
        })

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        raise NotImplementedError('do not call the abstract model!')

    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        raise NotImplementedError('do not call the abstract model!')



from ..types import AgentRun

from typing import List, Any


class NoResult(Exception):
    pass


def _extract_last_result_from_run(run: AgentRun) -> Any:
    if run['type'] == 'extraction':
        return run['result']
    elif run['type'] == 'messages':
        raise NoResult()
    elif run['type'] == 'chain' or run['type'] == 'parallel':
        for r in reversed(run['runs']):
            try:
                return _extract_last_result_from_run(r)
            except NoResult:
                pass  # keep iterate...
        raise NoResult()
    else:
        raise RuntimeError('unreachable')


def extract_is_trivial(
    extraction_run: AgentRun,
) -> bool:
    result = _extract_last_result_from_run(extraction_run)
    extraction_type = type(result)

    if hasattr(result, 'is_trivial'):
        is_trivial = result.is_trivial
    elif isinstance(result, dict) and ('is_trivial' in result):
        is_trivial = result['is_trivial']
    else:
        raise ValueError(f"{extraction_type} must have an `is_trivial` property or key to work in this planner")

    assert isinstance(is_trivial, bool), f"{extraction_type}.is_trivial must be a boolean; instead it is a {type(is_trivial)}"
    return is_trivial


def extract_subtask_strings(
    extraction_run: AgentRun,
) -> List[str]:
    result = _extract_last_result_from_run(extraction_run)
    extraction_type = type(result)

    if hasattr(result, 'subtasks'):
        subtasks = result.subtasks
    elif isinstance(result, dict) and ('subtasks' in result):
        subtasks = result['subtasks']
    else:
        raise ValueError(f"{extraction_type} must have an `subtasks` property or key to work in this planner")

    assert isinstance(subtasks, list), f"{extraction_type}.subtasks must be a list; instead it is a {type(subtasks)}"
    for i, s in enumerate(subtasks):
        assert isinstance(s, str), f"subtasks[{i}] must be a str; instead it is a {type(s)}"
    return subtasks


def extract_task_statement(
    extraction_run: AgentRun,
) -> str:
    result = _extract_last_result_from_run(extraction_run)
    extraction_type = type(result)

    if hasattr(result, 'task_statement'):
        task_statement = result.task_statement
    elif isinstance(result, dict) and ('task_statement' in result):
        task_statement = result['task_statement']
    else:
        raise ValueError(f"{extraction_type} must have an `task_statement` property or key to work in this planner")

    assert isinstance(task_statement, str), f"{extraction_type}.task_statement must be a str; instead it is a {type(task_statement)}"
    return task_statement



import pytest

from lasagna.agent_util import (
    make_model_binder,
    noop_callback,
    human_input,
    extract_last_message,
    override_system_prompt,
)

from lasagna.types import AgentRun

from lasagna.planner.debug_model import DebugModel

from lasagna.planner.agent import (
    build_default_planning_agent,
    _minimal_copy_lrpa_aware,
    build_message_extractor_lrpa_aware,
)

import copy
from typing import List


def _count_runs(run: AgentRun) -> int:
    subruns = 0
    if run['type'] == 'chain' or run['type'] == 'parallel':
        subruns = sum(_count_runs(r) for r in run['runs'])
    return subruns + 1


def _assert_recursive_structure(run: AgentRun) -> None:
    assert run['agent'] == 'planning_agent_chain'
    assert run['type'] == 'chain'

    runs = run['runs']

    if len(runs) == 2:
        assert runs[0]['agent'] == 'extraction_agent'
        assert runs[1]['agent'] == 'answer_chain'

    elif len(runs) == 3:
        assert runs[0]['agent'] == 'extraction_agent'
        assert runs[1]['agent'] == 'subtask_chain_of_chains'
        assert runs[2]['agent'] == 'answer_chain'

        subtask_chain_of_chains = runs[1]
        assert subtask_chain_of_chains['type'] == 'chain'

        for subtask_chain in subtask_chain_of_chains['runs']:
            assert subtask_chain['agent'] == 'subtask_chain'
            assert subtask_chain['type'] == 'chain'
            subtask_runs = subtask_chain['runs']
            assert len(subtask_runs) == 2
            assert subtask_runs[0]['agent'] == 'subtask_input_generator'
            _assert_recursive_structure(subtask_runs[1])

    else:
        assert False, f"why is len(runs) == {len(runs)}?"


@pytest.mark.asyncio
async def test_agent():
    binder_seed_0 = make_model_binder(DebugModel, 'NA', seed = 0)
    binder_seed_1 = make_model_binder(DebugModel, 'NA', seed = 1)

    planning_agent = build_default_planning_agent(binder = binder_seed_0)
    run = await planning_agent(noop_callback, human_input('hi'))
    assert _count_runs(run) == 47         # <-- if fails, did self.rand in the DebugModel change?
    _assert_recursive_structure(run)
    assert extract_last_message(
        run,
        from_tools = False,
        from_extraction = False,
    ).get('text') == 'The answer is: 23'  # <-- also depends on self.rand being reproducible

    planning_agent = build_default_planning_agent(binder = binder_seed_1)
    run = await planning_agent(noop_callback, human_input('hi'))
    assert _count_runs(run) == 6          # <-- if fails, did self.rand in the DebugModel change?
    _assert_recursive_structure(run)
    assert extract_last_message(
        run,
        from_tools = False,
        from_extraction = False,
    ).get('text') == 'The answer is: 3'   # <-- also depends on self.rand being reproducible


def test_minimal_copy_lrpa_aware():
    orig: List[AgentRun] = human_input('hi')
    orig_copy = copy.deepcopy(orig)
    modified_copy, inner_run, depth = _minimal_copy_lrpa_aware(orig)
    assert orig == orig_copy, "ensure pure function"
    assert depth == 0
    assert modified_copy == [
        {
            'agent': 'human_input',
            'type': 'messages',
            'messages': [
                {
                    'role': 'human',
                    'text': 'hi',
                },
            ],
        },
        {
            'agent': 'planning_agent_chain',
            'type': 'chain',
            'runs': [],
        },
    ]
    assert inner_run is modified_copy[1]

    orig: List[AgentRun] = [
        {
            'agent': 'human_input',
            'type': 'messages',
            'messages': [
                {
                    'role': 'human',
                    'text': 'hi',
                },
            ],
        },
        {
            'agent': 'planning_agent_chain',
            'type': 'chain',
            'runs': [
                {
                    'agent': 'extraction_agent',
                    'type': 'extraction',
                    'messages': [],
                    'result': {},
                },
                {
                    'agent': 'subtask_chain_of_chains',
                    'type': 'chain',
                    'runs': [
                        {
                            'agent': 'subtask_chain',
                            'type': 'chain',
                            'runs': [
                                {
                                    'agent': 'subtask_input_generator',
                                    'type': 'messages',
                                    'messages': [
                                        {
                                            'role': 'human',
                                            'text': '...',
                                        },
                                    ],
                                },
                            ],
                        },
                    ],
                },
            ],
        },
    ]
    orig_copy = copy.deepcopy(orig)
    modified_copy, inner_run, depth = _minimal_copy_lrpa_aware(orig)
    assert orig == orig_copy, "ensure pure function"
    assert depth == 1
    assert modified_copy == [
        {
            'agent': 'human_input',
            'type': 'messages',
            'messages': [
                {
                    'role': 'human',
                    'text': 'hi',
                },
            ],
        },
        {
            'agent': 'planning_agent_chain',
            'type': 'chain',
            'runs': [
                {
                    'agent': 'extraction_agent',
                    'type': 'extraction',
                    'messages': [],
                    'result': {},
                },
                {
                    'agent': 'subtask_chain_of_chains',
                    'type': 'chain',
                    'runs': [
                        {
                            'agent': 'subtask_chain',
                            'type': 'chain',
                            'runs': [
                                {
                                    'agent': 'subtask_input_generator',
                                    'type': 'messages',
                                    'messages': [
                                        {
                                            'role': 'human',
                                            'text': '...',
                                        },
                                    ],
                                },
                                {
                                    'agent': 'planning_agent_chain',
                                    'type': 'chain',
                                    'runs': [],
                                },
                            ],
                        },
                    ],
                },
            ],
        },
    ]
    assert inner_run is modified_copy[1]['runs'][1]['runs'][0]['runs'][1]  # type: ignore


def test_build_message_extractor_lrpa_aware():
    runs: List[AgentRun] = [
        {
            'agent': 'human_input',
            'type': 'messages',
            'messages': [
                {
                    'role': 'human',
                    'text': 'hi',
                },
            ],
        },
        {
            'agent': 'planning_agent_chain',
            'type': 'chain',
            'runs': [
                {
                    'agent': 'extraction_agent',
                    'type': 'extraction',
                    'messages': [],
                    'result': {'extraction_level': 0},
                },
                {
                    'agent': 'subtask_chain_of_chains',
                    'type': 'chain',
                    'runs': [
                        {
                            'agent': 'subtask_chain',
                            'type': 'chain',
                            'runs': [
                                {
                                    'agent': 'subtask_input_generator',
                                    'type': 'messages',
                                    'messages': [
                                        {
                                            'role': 'human',
                                            'text': '... go deeper on task 1',
                                        },
                                    ],
                                },
                                {
                                    'agent': 'planning_agent_chain',
                                    'type': 'chain',
                                    'runs': [
                                        {
                                            'agent': 'extraction_agent',
                                            'type': 'extraction',
                                            'messages': [],
                                            'result': {'extraction_level': 1},
                                        },
                                        {
                                            'agent': 'answer_chain',
                                            'type': 'chain',
                                            'runs': [
                                                {
                                                    'agent': 'answer_input_chain',
                                                    'type': 'chain',
                                                    'runs': [
                                                        {
                                                            'agent': 'answer_input_prompt',
                                                            'type': 'messages',
                                                            'messages': [
                                                                {
                                                                    'role': 'human',
                                                                    'text': '... make your final answer',
                                                                },
                                                            ],
                                                        },
                                                    ],
                                                },
                                                {
                                                    'agent': 'answer_agent',
                                                    'type': 'messages',
                                                    'messages': [
                                                        {
                                                            'role': 'ai',
                                                            'text': 'it is 47',
                                                        },
                                                        {
                                                            'role': 'human',
                                                            'text': 'really?',
                                                        },
                                                        {
                                                            'role': 'ai',
                                                            'text': 'yes',
                                                        },
                                                    ],
                                                },
                                            ],
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            'agent': 'subtask_chain',
                            'type': 'chain',
                            'runs': [
                                {
                                    'agent': 'subtask_input_generator',
                                    'type': 'messages',
                                    'messages': [
                                        {
                                            'role': 'human',
                                            'text': '... go deeper on task 2',
                                        },
                                    ],
                                },
                                {
                                    'agent': 'planning_agent_chain',
                                    'type': 'chain',
                                    'runs': [
                                        {
                                            'agent': 'extraction_agent',
                                            'type': 'extraction',
                                            'messages': [],
                                            'result': {'extraction_level': 2},
                                        },
                                        {
                                            'agent': 'answer_chain',
                                            'type': 'chain',
                                            'runs': [
                                                {
                                                    'agent': 'answer_input_chain',
                                                    'type': 'chain',
                                                    'runs': [
                                                        {
                                                            'agent': 'answer_input_prompt',
                                                            'type': 'messages',
                                                            'messages': [
                                                                {
                                                                    'role': 'human',
                                                                    'text': '... make your final answer now',
                                                                },
                                                            ],
                                                        },
                                                    ],
                                                },
                                                {
                                                    'agent': 'answer_agent',
                                                    'type': 'messages',
                                                    'messages': [
                                                        {
                                                            'role': 'ai',
                                                            'text': 'it is 48',
                                                        },
                                                        {
                                                            'role': 'human',
                                                            'text': 'really?',
                                                        },
                                                        {
                                                            'role': 'ai',
                                                            'text': 'yes!',
                                                        },
                                                    ],
                                                },
                                            ],
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            'agent': 'subtask_chain',
                            'type': 'chain',
                            'runs': [
                                {
                                    'agent': 'subtask_input_generator',
                                    'type': 'messages',
                                    'messages': [
                                        {
                                            'role': 'human',
                                            'text': '... go deeper on task 3',
                                        },
                                    ],
                                },
                                {
                                    'agent': 'planning_agent_chain',
                                    'type': 'chain',
                                    'runs': [
                                        {
                                            'agent': 'extraction_agent',
                                            'type': 'extraction',
                                            'messages': [],
                                            'result': {'extraction_level': 3},
                                        },
                                        {
                                            'agent': 'answer_chain',
                                            'type': 'chain',
                                            'runs': [
                                                {
                                                    'agent': 'answer_input_chain',
                                                    'type': 'chain',
                                                    'runs': [
                                                        {
                                                            'agent': 'answer_input_prompt',
                                                            'type': 'messages',
                                                            'messages': [
                                                                {
                                                                    'role': 'human',
                                                                    'text': '... make your final answer now again',
                                                                },
                                                            ],
                                                        },
                                                    ],
                                                },
                                            ],
                                        },
                                    ],
                                },
                            ],
                        },
                    ],
                },
            ],
        },
    ]
    runs_copy = copy.deepcopy(runs)
    messages_no_system_prompt = build_message_extractor_lrpa_aware(
        system_prompt_override=None,
    )(runs)
    messages_with_system_prompt = build_message_extractor_lrpa_aware(
        system_prompt_override='sys',
    )(runs)
    assert runs == runs_copy, "ensure pure function"
    assert override_system_prompt(messages_no_system_prompt, 'sys') == messages_with_system_prompt
    assert messages_no_system_prompt == [
        {
            'role': 'human',
            'text': 'hi',
        },
        {
            'role': 'ai',
            'text': '{"extraction_level": 0}',
        },
        {
            'role': 'human',
            'text': '... go deeper on task 1',
        },
        {
            'role': 'ai',
            'text': '{"extraction_level": 1}',
        },
        {
            'role': 'human',
            'text': '... make your final answer',
        },
        {
            'role': 'ai',
            'text': 'it is 47',
        },
        {
            'role': 'human',
            'text': 'really?',
        },
        {
            'role': 'ai',
            'text': 'yes',
        },
        {
            'role': 'human',
            'text': '... go deeper on task 2',
        },
        {
            'role': 'ai',
            'text': '{"extraction_level": 2}',
        },
        {
            'role': 'human',
            'text': '... make your final answer now',
        },
        {
            'role': 'ai',
            'text': 'it is 48',
        },
        {
            'role': 'human',
            'text': 'really?',
        },
        {
            'role': 'ai',
            'text': 'yes!',
        },
        {
            'role': 'human',
            'text': '... go deeper on task 3',
        },
        {
            'role': 'ai',
            'text': '{"extraction_level": 3}',
        },
        {
            'role': 'human',
            'text': '... make your final answer now again',
        },
    ]

    # Now, if we append the answer to the top-level planning_agent_chain,
    # then all the stuff inside will collapse.
    top_level_planner_run = runs[1]
    assert top_level_planner_run['agent'] == 'planning_agent_chain'
    assert top_level_planner_run['type'] == 'chain'
    top_level_planner_run['runs'].append({
        'agent': 'answer_chain',
        'type': 'chain',
        'runs': [
            {
                'agent': 'answer_input_chain',
                'type': 'chain',
                'runs': [
                    {
                        'agent': 'answer_input_prompt',
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'human',
                                'text': 'FINAL ANSWER?',
                            },
                        ],
                    },
                ],
            },
            {
                'agent': 'answer_agent',
                'type': 'messages',
                'messages': [
                    {
                        'role': 'ai',
                        'text': 'answer is 7',
                    },
                    {
                        'role': 'human',
                        'text': 'no?',
                    },
                    {
                        'role': 'ai',
                        'text': 'sure is',
                    },
                ],
            },
        ],
    })
    runs_copy = copy.deepcopy(runs)
    messages_no_system_prompt = build_message_extractor_lrpa_aware(
        system_prompt_override=None,
    )(runs)
    messages_with_system_prompt = build_message_extractor_lrpa_aware(
        system_prompt_override='sys',
    )(runs)
    assert runs == runs_copy, "ensure pure function"
    assert override_system_prompt(messages_no_system_prompt, 'sys') == messages_with_system_prompt
    assert messages_no_system_prompt == [
        {
            'role': 'human',
            'text': 'hi',
        },
        {
            'role': 'ai',
            'text': '{"extraction_level": 0}',
        },
        {
            'role': 'human',
            'text': 'FINAL ANSWER?',
        },
        {
            'role': 'ai',
            'text': 'answer is 7',
        },
        {
            'role': 'human',
            'text': 'no?',
        },
        {
            'role': 'ai',
            'text': 'sure is',
        },
    ]



import asyncio
import functools
import time

from typing import (
    Optional, Union,
    TypeVar, Callable, Awaitable,
)

from typing_extensions import ParamSpec


P = ParamSpec('P')  # <-- for generic programming
T = TypeVar('T')    # <-- for generic programming


def async_throttle(
    max_concurrent: Optional[int] = None,
    max_per_second: Optional[float] = None,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """
    Decorator for async functions to throttle their invocations to stay
    under the specified `max_concurrent` and/or `max_per_second`.
    """
    if max_concurrent is None and max_per_second is None:
        raise ValueError(f"you must specify *at least one* of `max_concurrent` or `max_per_second`")

    elif max_concurrent is not None and max_per_second is not None:
        def decorator(wrapped: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            tc = _async_throttle_concurrent(max_concurrent)
            tr = _async_throttle_rate(max_per_second)
            return tc(tr(wrapped))  # <-- this is the correct order to apply the decorators
        return decorator

    elif max_concurrent is not None:
        return _async_throttle_concurrent(max_concurrent)

    elif max_per_second is not None:
        return _async_throttle_rate(max_per_second)

    else:
        raise RuntimeError("unreachable")


def _async_throttle_concurrent(
    max_concurrent: int,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    def decorator(wrapped: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        sem = asyncio.Semaphore(max_concurrent)

        @functools.wraps(wrapped)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            async with sem:
                return await wrapped(*args, **kwargs)

        return wrapper

    return decorator


def _async_throttle_rate(
    max_per_second: float,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    def decorator(wrapped: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        period = 1.0 / max_per_second
        last_call: Union[float, None] = None

        @functools.wraps(wrapped)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            nonlocal last_call
            now = time.time()
            if last_call is None:
                last_call = now
                delay = 0.0
            else:
                this_call = last_call + period
                delay = max(0.0, this_call - now)
                last_call = max(this_call, now)
            if delay > 0.0:
                await asyncio.sleep(delay)
            return await wrapped(*args, **kwargs)

        return wrapper

    return decorator



from ..types import AgentRun
from ..agent_util import flat_messages
from ..util import get_name
from ..tools_util import (
    convert_to_json_schema,
    get_tool_params,
)

from .util import extract_subtask_strings, extract_task_statement

from pydantic import BaseModel, Field

import json
from typing import List, Dict, AsyncIterable, Callable


PLANNING_EXTRACTION_AGENT_SYSTEM_PROMPT = """
You plan the execution of a task by:
- analyzing it,
- determining its complexities, and
- (if required) breaking it up into subtasks.
""".strip()


class PlanOutput(BaseModel):
    thoughts: str = Field(description="your free-form thoughts about the user's most recent prompt")
    task_statement: str = Field(description="a concisely-stated task statement, fully describing what the user wants in their most recent prompt")
    is_trivial: bool = Field(description="`true` if this task has a trivial solution (such as recalling a simple fact, or making simple tool calls); `false` if this task needs to be broken up into subtasks, especially when those subtasks are exploratory in their nature")
    subtasks: List[str] = Field(description="an ordered list of subtasks that need to be performed in order to complete the task statement (if not trivial); else an empty list (if trivial)")


async def subtask_input_generator(
    extraction_run: AgentRun,
) -> AsyncIterable[List[AgentRun]]:
    subtask_strings = extract_subtask_strings(extraction_run)
    for i, subtask_string in enumerate(subtask_strings):
        yield [
            flat_messages('subtask_input_generator', [{
                'role': 'human',
                'text': f'Now work on subtask #{i+1}: {subtask_string}',
            }]),
        ]


async def answer_input_prompt(
    extraction_run: AgentRun,
) -> List[AgentRun]:
    task_statement = extract_task_statement(extraction_run)
    return [
        flat_messages('answer_input_prompt', [{
            'role': 'human',
            'text': f"Provide your final answer to the task: {task_statement}",
        }]),
    ]


def append_tool_spec_to_prompt(
    prompt: str,
    tools: List[Callable],
) -> str:
    if not tools:
        return prompt

    tool_specs: List[Dict] = []

    for tool in tools:
        description, params = get_tool_params(tool)
        tool_specs.append({
            'name': get_name(tool),
            'description': description,
            'input_schema': convert_to_json_schema(params),
        })

    sections = [
        prompt,
        "While planning, you may refer to the following tools:",
        json.dumps(tool_specs, indent=2),
    ]

    return '\n\n'.join(sections)



import pytest
from pydantic import BaseModel, Field

from lasagna import (
    make_model_binder,
    easy_ask,
    easy_extract,
)
from lasagna.mock_provider import (
    MockProvider,
)


@pytest.mark.asyncio
async def test_easy_ask():
    result: str = await easy_ask(
        binder = make_model_binder(MockProvider, 'fake_model'),
        prompt = 'Hello, world!',
        system_prompt = 'Say hello world.',
    )
    assert result == "\n".join([
        "Say hello world.",
        "Hello, world!",
        "model: fake_model",
    ])


@pytest.mark.asyncio
async def test_easy_extract():
    class MyModel(BaseModel):
        name: str = Field(description='Make up a name of the person')
        age: int = Field(description='Make up an age of the person')

    result = await easy_extract(
        binder = make_model_binder(MockProvider, 'fake_model', **dict(
            name = 'John Doe',
            age = 30,
        )),
        prompt = 'Hello, world!',
        system_prompt = 'You are making stuff up.',
        extraction_type = MyModel,
    )
    assert isinstance(result, MyModel), f"Expected MyModel type, got {type(result)}"
    assert result.name == 'John Doe'
    assert result.age == 30


@pytest.mark.asyncio
async def test_easy_ask_with_streaming():
    streaming_output: str = ""
    async def streaming_callback(text: str) -> None:
        nonlocal streaming_output
        streaming_output += text

    result: str = await easy_ask(
        binder = make_model_binder(MockProvider, 'fake_model'),
        prompt = 'Hello, world!',
        system_prompt = 'Say hello world.',
        streaming_callback = streaming_callback,
    )
    assert len(streaming_output) > 0, \
        f'streaming_output: {streaming_output}'

